#include "internal.h"
#include "midiseq.h"
#include <stdlib.h>

static struct jfaud_init_params *params = NULL;

static int cmdlens[] = {
	2 /*noteoff 0x80*/, 2 /*noteon 0x90*/, 2 /*aftertouch 0xA0*/, 2 /*control 0xB0*/,
	1 /*program 0xC0*/, 1 /*channelaftertouch 0xD0*/, 2 /*pitchwheel 0xE0*/, 0 /*syscommon 0xF0*/
};


unsigned long midifile::getmidinum()
{
	long val = 0;
	unsigned char by;
	do {
		by = getbyte();
		val <<= 7;
		val |= (by&0x7f);
	} while (by&0x80);
	return val;
}

unsigned long midifile::getbignum(unsigned char *ptr, int size)
{
	unsigned long val = 0;
	if (size > 4) size = 4;
	while (size--) {
		val <<= 8;
		val |= *(ptr++);
	}
	return val;
}

int midifile::scan(unsigned char device)
{
	int i;
	char foundincl;
	long deltatime;
	unsigned char command, parm1, parm2;

	for (i=0; i<numtracks; i++) {
		foundincl = 0;
		command = 0;
		selecttrack(i);
		rewind();
		clrflag(255);
		while (!eot()) {
			deltatime = getmidinum();
			if (peekbyte(0) & 0x80) command = getbyte();
			else if (!command)
				while (!((command = getbyte()) & 0x80)) ;

			if ((command&0xF0) == 0xB0) {	// controller change
				parm1 = getbyte();
				parm2 = getbyte();
				switch (parm1) {
					case 110:	// EMIDI include
						if (parm2 == 127 || parm2 == device) clrflag(IgnoreTrack);
						else if (!foundincl) setflag(IgnoreTrack);
						foundincl = 1;
						break;
						
					case 111:	// EMIDI exclude
						if (parm2 == 127 || parm2 == device)
							setflag(IgnoreTrack);
						break;

					case 112:	// EMIDI program
						setflag(IgnorePrograms);
						break;

					case 113:	// EMIDI volume
						setflag(IgnoreVolumes);
						break;
				}
			} else if ((command&0xF0) == 0xF0) {	// meta event
				switch (command) {
					case 0xF0:
						while (getbyte() != 0xF7) ;
						break;
					case 0xFF:
						getbyte();
						for (parm2=getbyte(); parm2>0; parm2--) getbyte();
						break;
					case 0xF2:
						getbyte();
					case 0xF1:
					case 0xF3:
						getbyte();
					default:
						break;
				}
				command = 0;
			} else {
				for (parm2=cmdlens[ (command>>4) & 7 ]; parm2>0; parm2--)
					getbyte();
			}
		}
	}
	return 0;
}

midifile::midifile()
{
	if (!params) params = jfaud_getparams();
	
	tracklength = NULL;
	trackpos = NULL;
	trackflags = NULL;
	tracks = NULL;
	curtrack = -1;
}

midifile::~midifile()
{
	if (tracklength) free(tracklength);
	if (trackpos) free(trackpos);
	if (trackflags) free(trackflags);
	if (tracks) {
		int i;
		for (i=0;i<numtracks;i++)
			if (tracks[i])
				free(tracks[i]);
		free(tracks);
	}
}

int midifile::loadfile(int fh)	// -1 = error, -2 = not midi
{
	int i;
	unsigned long chunksz;
	unsigned char headerbuf[2*8];

	if (params->call_read(fh, headerbuf, 4*2) != 4*2) {
		debuglog(params->call_logmsg,"midiseq: Truncated header.");
		return -1;
	}

	if (headerbuf[0] != 'M' || headerbuf[1] != 'T' || headerbuf[2] != 'h' || headerbuf[3] != 'd') {
		debuglog(params->call_logmsg,"midiseq: File doesn't look like MIDI.");
		return -2;
	}

	if ((chunksz=getbignum(&headerbuf[4], 4)) < 6) {
		debuglog(params->call_logmsg,"midiseq: Header is too short.");
		return -1;
	}

	if (params->call_read(fh, headerbuf, 2*3) != 2*3) {
		debuglog(params->call_logmsg,"midiseq: Truncated header.");
		return -1;
	}

	format = getbignum(&headerbuf[0], 2);
	numtracks = getbignum(&headerbuf[2], 2);
	ppqn = getbignum(&headerbuf[4], 2);

	debuglog(params->call_logmsg,jsprintf("midiseq: Format: %d, Tracks: %d, Ticks/quarter note: %d",
			format,numtracks,ppqn));

	tracklength = (unsigned long *)calloc(numtracks, sizeof(unsigned long));
	tracks      = (unsigned char **)calloc(numtracks, sizeof(unsigned char *));
	trackpos    = (unsigned char **)calloc(numtracks, sizeof(unsigned char *));
	trackflags  = (unsigned char *)calloc(numtracks, sizeof(unsigned char));
	if (!tracklength || !tracks || !trackpos || !trackflags) return -1;
	
	for (i=0; i<numtracks; i++) {
		if (params->call_read(fh, headerbuf, 4*2) != 4*2) {
			debuglog(params->call_logmsg,"midiseq: Truncated track.");
			return -1;
		}

		if (headerbuf[0] != 'M' || headerbuf[1] != 'T' || headerbuf[2] != 'r' || headerbuf[3] != 'k') {
			debuglog(params->call_logmsg,"midiseq: Track header invalid.");
			return -1;
		}

		tracklength[i] = getbignum(&headerbuf[4], 4);
		tracks[i] = (unsigned char *)calloc(1, tracklength[i]);
		if (!tracks[i]) return -1;

		debuglog(params->call_logmsg,jsprintf("midiseq: Track %d: size=%d",i,tracklength[i]));

		if (params->call_read(fh, tracks[i], tracklength[i]) != tracklength[i]) {
			debuglog(params->call_logmsg,"midiseq: Track data truncated.");
			return -1;
		}
	}

	return 0;
}

int midifile::setdevice(unsigned char dev)
{
	if (dev > 127) dev = 127;
	return scan(dev);
}



midiseq::midiseq() : midifile()
{
	seqtrack = NULL;
	lastcommand = 0;
	infiniteloop = 0;
}

midiseq::~midiseq()
{
	if (seqtrack) free(seqtrack);
}

int midiseq::fetchevt(int track)
{
	int cmd,len;
	
	selecttrack(track);
	if (eot()) return 1;

	seqtrack[track].delta = getmidinum();
	if (peekbyte(0) & 0x80) seqtrack[track].command = getbyte();
	cmd = seqtrack[track].command;
	len = cmdlens[ (cmd>>4) & 7 ];
	seqtrack[track].evtptr = getptr();
	seqtrack[track].evtlen = len;

	if (len == 0) {	// system common event
		switch (cmd) {
			case 0xF0:
				// sysex
				do seqtrack[track].evtlen++; while (getbyte() != 0xF7);
				break;
			case 0xFF:
				// meta
				cmd = getbyte();
				len = getbyte();
				seqtrack[track].evtlen += len + 2;
				while (len--) getbyte();
				break;
			case 0xF2:
				// two-byte messages
				seqtrack[track].evtlen++; getbyte();
			case 0xF1:
			case 0xF3:
				// one-byte messages
				seqtrack[track].evtlen++; getbyte();
			default:
				// command-byte-only messages
				break;
		}
	} else {
		while (len--) getbyte();
	}

	return 0;
}

int midiseq::setupsequencer(unsigned char dev, int infinite)
{
	int i;
	
	if (setdevice(dev)) return -1;

	infiniteloop = infinite;

	if (seqtrack) free(seqtrack);
	seqtrack = (struct trackdef *)calloc(getnumtracks(), sizeof(struct trackdef));

	for (i=0;i<getnumtracks();i++) {
		selecttrack(i);
		if (testflag(IgnoreTrack)) { seqtrack[i].inactive = 1; continue; }
		rewind();
		if (fetchevt(i)) seqtrack[i].inactive = 1;
	}

	return 0;
}

int midiseq::getevent(unsigned long *delta, unsigned char *command, unsigned char const **evtdata, unsigned long *len)
{
	// returns 0 if a message was returned, 1 if the end of the song was reached, or -1 on error

	unsigned long deltaacc = 0, mindelta = 0xfffffff;
	int mindeltatrk = -1, i, j;
	char gotevent = 0;

	static unsigned char evtbuf[3];

again:
	// find the track with the nearest event
	mindelta = 0xffffffff;
	mindeltatrk = -1;
	for (i=0;i<getnumtracks();i++) {
		if (seqtrack[i].inactive) continue;
		if (seqtrack[i].delta >= mindelta) continue;
		mindelta = seqtrack[i].delta;
		mindeltatrk = i;
	}
	if (mindeltatrk < 0) {
		if (!infiniteloop) return 1;
		for (i=0;i<getnumtracks();i++) {
			selecttrack(i);
			if (testflag(IgnoreTrack)) { seqtrack[i].inactive = 1; continue; }
			seqtrack[i].inactive = 0;
			rewind();
			if (fetchevt(i)) seqtrack[i].inactive = 1;
		}
		goto again;
	}

	// check out the event
	selecttrack(mindeltatrk);
	switch (seqtrack[mindeltatrk].command & 0xF0) {
		case 0xC0:	// program
			if (testflag(IgnorePrograms))
				break;
		case 0x80:	// note off
		case 0x90:	// note on
		case 0xA0:	// aftertouch
		case 0xD0:	// channel aftertouch
		case 0xE0:	// pitch wheel
			gotevent = 1;
			break;
		case 0xF0:	// system common event
			// we keep tempo meta events so they can be interpreted, but all other metas get swallowed
			if (seqtrack[mindeltatrk].command == 0xFF && seqtrack[mindeltatrk].evtptr[0] == 0x51) gotevent = 1;
			else if (seqtrack[mindeltatrk].command == 0xFF && seqtrack[mindeltatrk].evtptr[0] == 0x2F)
				seqtrack[i].inactive = 1;
			break;
		case 0xB0:	// control
			switch (seqtrack[mindeltatrk].evtptr[0]) {
				case 110:	// EMIDI include
				case 111:	// EMIDI exclude
				case 114:	// EMIDI context start
				case 115:	// EMIDI context end
					break;
				case 113:	// EMIDI volume
					evtbuf[0] = 7;		// cook up a normal volume change event
					evtbuf[1] = seqtrack[mindeltatrk].evtptr[1];
					*command = seqtrack[mindeltatrk].command;
					*evtdata = evtbuf;
					*len     = 2;
					gotevent = 2;
					break;
				case 112:	// EMIDI program
					*command  = 0xC0 | (seqtrack[mindeltatrk].command & 15);	// cook up a normal
					evtbuf[0] = seqtrack[mindeltatrk].evtptr[1] & 0x7F;		// program change event
					*evtdata = evtbuf;
					*len     = 1;
					gotevent = 2;
					break;
				case 116:	// EMIDI loop begin
				case 118:	// EMIDI song loop begin
					if (seqtrack[mindeltatrk].evtptr[1] == 0) {
						j = -1;
					} else {
						j = seqtrack[mindeltatrk].evtptr[1];
						debuglog(params->call_logmsg,jsprintf("midiseq: Track %d: Entering %d-cycle loop",mindeltatrk,j));
					}
					
					if (seqtrack[mindeltatrk].evtptr[0] == 118) {
						// song-wide loop
						for (i=0;i<getnumtracks();i++) {
							seqtrack[i].loopcount = j;
							selecttrack(i);
							seqtrack[i].loopstart = getpos();
						}
					} else {
						seqtrack[mindeltatrk].loopcount = j;
						seqtrack[mindeltatrk].loopstart = getpos();
					}
					break;
				case 117:	// EMIDI loop end
				case 119:	// EMIDI song loop end
					if (seqtrack[mindeltatrk].evtptr[1] != 127) {
						break;
					}
					if (!infiniteloop && seqtrack[mindeltatrk].loopcount < 0) {
						break;
					}

					if (seqtrack[mindeltatrk].evtptr[0] == 118) {
						// song-wide loop
						for (i=0;i<getnumtracks();i++) {
							selecttrack(i);
							seqtrack[i].loopend = getpos();
						}
						selecttrack(mindeltatrk);
					}
					
					if (seqtrack[mindeltatrk].loopcount) {
						if (seqtrack[mindeltatrk].loopcount > 0) {
							seqtrack[mindeltatrk].loopcount--;
							debuglog(params->call_logmsg,jsprintf("midiseq: Track %d: %d to go...",mindeltatrk,seqtrack[mindeltatrk].loopcount));
						}
						setpos(seqtrack[mindeltatrk].loopstart);
					}
					break;
				case 7:		// volume change
					if (testflag(IgnoreVolumes))
						break;
				default:
					gotevent = 1;
					break;
			}
			break;
	}

	if (gotevent) {
		if (gotevent == 1) {
			*command = seqtrack[mindeltatrk].command;
			*evtdata = seqtrack[mindeltatrk].evtptr;
			*len     = seqtrack[mindeltatrk].evtlen;
		}
		lastcommand = *command;
		*delta = deltaacc + mindelta;
	}
	
	// move all events forward and get a new one for the track we just ate
	for (i=0;i<getnumtracks();i++) {
		if (seqtrack[i].inactive) continue;
		if (i == mindeltatrk) {
			if (fetchevt(i)) {
				seqtrack[i].inactive = 1;
			}
		} else {
			seqtrack[i].delta -= mindelta;
		}
	}

	if (gotevent) {
		return 0;
	} else {
		deltaacc += mindelta;
		goto again;
	}
	
	return -1;
}

