#include "internal.h"

#include <string.h>
#include <stdlib.h>

static struct jfaud_init_params *params;

#define maxopenfiles	64
static jfaud_file files[maxopenfiles];

static char initialised = 0;


jfauderr jfaudfile_init(void)
{
	if (initialised) return jfauderr_alreadyinited;
	
	params = jfaud_getparams();
	if (!params) return jfauderr_genericerror;

	memset(files, 0, sizeof(files));
	
	initialised = 1;
	
	return jfauderr_ok;
}

jfauderr jfaudfile_uninit(void)
{
	int i;

	if (!initialised) return jfauderr_notinited;

	for (i=maxopenfiles-1;i>=0;i--) {
		if (files[i].ref == 0) continue;
	
		files[i].ref = 1;
		jfaudfile_close(&files[i]);
	}

	initialised = 0;

	return jfauderr_ok;
}

jfauderr jfaudfile_open(const char *fn, jfaud_file **file)
{
	int i;
	int firstfree = -1;
	unsigned long fnhash;
	char *basefn, *subfn;
	jfauderr err,rerr;

	if (!initialised) return jfauderr_notinited;

	fnhash = jfaud_hashfn(fn);
	debuglog(params->call_logmsg,jsprintf("fnhash: \"%s\"=%x",fn,fnhash));

	for (i=maxopenfiles-1; i>=0; i--) {
		if (files[i].ref == 0) {
			firstfree = i;
			break;
		}
	}

	if (firstfree == -1) {
		*file = NULL;
		//jfaudfile_dumpstatus();
		return jfauderr_openerror;
	}

	*file = &files[firstfree];

	basefn = strdup(fn);
	subfn = strrchr(basefn,'|');
	if (subfn) { *subfn = 0; subfn++; }

	files[firstfree].type = jfaudfile_unidentified;
	files[firstfree].readerinfo = NULL;
	files[firstfree].ref = 0;
	files[firstfree].ident = fnhash;
	files[firstfree].fh = params->call_open(basefn);
	if (files[firstfree].fh == -1) {
		free(basefn);
		//jfaudfile_dumpstatus();
		return jfauderr_openerror;
	}

	files[firstfree].ref = 1;

	// ask each of the file readers whether it knows what the file is
	rerr = jfauderr_unrecognised;
	for (i=0; i<numfilereaders; i++) {
		if (!filereaderspresent[i]) continue;	// reader is disabled
		err = filereaders[i]->testfile(&files[firstfree], subfn);
		if (err == jfauderr_unrecognised) continue;
		else if (err == jfauderr_ok) {
			// the file has been identified and is ready to read
			free(basefn);
			return jfauderr_ok;
		} else {
			rerr = err;
			break;
		}
	}

	params->call_close(files[firstfree].fh);
	free(basefn);

	return rerr;
}


jfauderr jfaudfile_close(jfaud_file *file)
{
	char noclose = 0;

	if (!initialised) return jfauderr_notinited;

	if (file->ref > 1) {
		file->ref--;
		return jfauderr_ok;
	} else if (file->ref == 0) {
		return jfauderr_ok;
	}
	
	if (file->fh == -1) return jfauderr_ok;

	if (file->type != jfaudfile_unidentified) {
		filereaders[file->type]->cleanup(file, &noclose);
	}

	if (!noclose) params->call_close(file->fh);
	file->fh = -1;
	file->type = jfaudfile_unidentified;
	file->readerinfo = NULL;
	file->ref = 0;

	return jfauderr_ok;
}


jfauderr jfaudfile_closehandle(jfaud_file *file)
{
	if (!initialised) return jfauderr_notinited;

	if (file->fh == -1) return jfauderr_ok;
	params->call_close(file->fh);
	file->fh = -1;
	return jfauderr_ok;
}


int jfaudfile_read(jfaud_file *file, void *buf, unsigned len)
{
	if (!file) return -1;
	if (file->fh < 0) return -1;

	return params->call_read(file->fh, buf, len);
}


int jfaudfile_seek(jfaud_file *file, int len, int whence)
{
	if (!file) return -1;
	if (file->fh < 0) return -1;

	return params->call_seek(file->fh, len, whence);
}


int jfaudfile_tell(jfaud_file *file)
{
	return params->call_tell(file->fh);
}


jfauderr jfaudfile_getpcmlength(jfaud_file *file, unsigned int *samples)
{
	if (!file || !samples) return jfauderr_genericerror;
	if (file->type == jfaudfile_unidentified) return jfauderr_unrecognised;
	if (!filereaders[ file->type ]->getpcmlength) return jfauderr_genericerror;
	return filereaders[ file->type ]->getpcmlength(file, samples);
}

jfauderr jfaudfile_getplaytime(jfaud_file *file, double *len)
{
	if (!file || !len) return jfauderr_genericerror;
	if (file->type == jfaudfile_unidentified) return jfauderr_unrecognised;
	if (!filereaders[ file->type ]->getplaytime) return jfauderr_genericerror;
	return filereaders[ file->type ]->getplaytime(file, len);
}

void jfaudfile_dumpstatus(void)
{
	int i;

	params->call_logmsg("jfaudfile_dumpstatus(): Beginning dump");
	for (i=0;i<maxopenfiles;i++) {
		params->call_logmsg(jsprintf(
			"File #%d: ref=%d ident=%X fh=%d type=%d",
			i, files[i].ref, (unsigned long)files[i].ident, files[i].fh, files[i].type
					));
	}
	params->call_logmsg("jfaudfile_dumpstatus(): End of dump");
}

