#include "internal.h"

#include <string.h>
#include <stdio.h>

#define KWV_SAMPLERATE 11025
#define KWV_CHANNELS 1
#define KWV_BYTESPERSAMP 1

struct kwvinfo {
	unsigned long audiostart;
	unsigned long audioleng;
	unsigned long repstart;
	unsigned long repleng;

	unsigned long audiopos;
};

static jfauderr init(void)
{
	return jfauderr_ok;
}

static jfauderr uninit(void)
{
	return jfauderr_ok;
}

static jfauderr testfile(jfaud_file *file, const char *sub)
{
	struct kwvinfo *kwvf;
	unsigned long version = 0;
	unsigned long numwaves = 0;
	unsigned long i;
	int j;

	unsigned long ast=0,alen=0,rst=0,rlen=0,l;
	char wvname[16], ch1, found=0;
	
	if (!sub || !sub[0]) return jfauderr_unrecognised;	// KWV files are a container, so a sub-file must be supplied
	if (strlen(sub) > 16) return jfauderr_unrecognised;	// name far too long

	jfaudfile_seek(file, 0, SEEK_SET);

	if (jfaudfile_read(file, &version, 4) != 4) return jfauderr_genericerror;
	if (version != 0) return jfauderr_unrecognised;

	if (jfaudfile_read(file, &numwaves, 4) != 4) return jfauderr_genericerror;

	while (numwaves-->0) {
		if (jfaudfile_read(file, wvname, 16) != 16) return jfauderr_genericerror;
		for (j=0;j<16 && sub[j];j++) {
			ch1 = sub[j];
			if (ch1 >= 97 && ch1 <= 123) ch1 -= 32;
			if (wvname[j] >= 97 && wvname[j] <= 123) wvname[j] -= 32;
			if (ch1 != wvname[j]) { j=16; break; }
		}
		if (j==16) {
			// invalid
			if (jfaudfile_read(file, &j, 4) != 4) return jfauderr_genericerror;
			ast += j;
			jfaudfile_seek(file, 4*3, SEEK_CUR);	// ignore the other fields
		} else {
			// found it
			if (jfaudfile_read(file, &alen, 4) != 4) return jfauderr_genericerror;
			if (jfaudfile_read(file, &rst, 4) != 4) return jfauderr_genericerror;
			if (jfaudfile_read(file, &rlen, 4) != 4) return jfauderr_genericerror;
			jfaudfile_seek(file, 4+(16+4*4)*numwaves, SEEK_CUR);	// skip the finetune and the remaining entries
			found = 1;
			break;
		}
	}

	if (!found) return jfauderr_genericerror;	// not found

	jfaudfile_seek(file, ast, SEEK_CUR);
	
	kwvf = (struct kwvinfo *)malloc(sizeof(struct kwvinfo));
	if (!kwvf) return jfauderr_outofmemory;

	kwvf->audiostart = jfaudfile_tell(file);
	kwvf->audioleng = alen;
	kwvf->repstart = rst;
	kwvf->repleng = rlen;
	kwvf->audiopos = 0;

	file->type = jfaudfile_kwv;
	file->readerinfo = (void *)kwvf;

	return jfauderr_ok;
}

static jfauderr cleanup(jfaud_file *file, char *noclose)
{
	if (file->readerinfo) free(file->readerinfo);
	file->readerinfo = NULL;
	*noclose = 0;

	return jfauderr_ok;
}

static jfauderr readpcmdata(jfaud_file *file, jfaud_buffer *data, int samples)
{
	struct kwvinfo *kwvf;
	int samps, sr;
	jfauderr err;

	if (file->type != jfaudfile_kwv) return jfauderr_unrecognised;

	kwvf = (struct kwvinfo *)file->readerinfo;

	samps = (kwvf->audioleng - kwvf->audiopos) / (KWV_CHANNELS * KWV_BYTESPERSAMP);
	if (samples < samps) samps = samples;

	if ((err = jfaudbuffer_alloc(data, KWV_SAMPLERATE, KWV_CHANNELS, KWV_BYTESPERSAMP*8, samps)) != jfauderr_ok)
		return err;

	sr = jfaudfile_read(file, data->data, samps * KWV_CHANNELS * KWV_BYTESPERSAMP);
	if (sr <= 0) {
		jfaudbuffer_free(data);
		if (sr < 0) return jfauderr_genericerror;
		return jfauderr_ok;
	}

	data->numsamples = sr / (KWV_CHANNELS * KWV_BYTESPERSAMP);
	kwvf->audiopos += sr;

	return jfauderr_ok;
}

static jfauderr readallpcmdata(jfaud_file *file, jfaud_buffer *data)
{
	struct kwvinfo *kwvf;
	int samps, sr;
	jfauderr err;

	if (file->type != jfaudfile_kwv) return jfauderr_unrecognised;

	kwvf = (struct kwvinfo *)file->readerinfo;

	samps = (kwvf->audioleng - kwvf->audiopos) / (KWV_CHANNELS * KWV_BYTESPERSAMP);

	if ((err = jfaudbuffer_alloc(data, KWV_SAMPLERATE, KWV_CHANNELS, KWV_BYTESPERSAMP*8, samps)) != jfauderr_ok)
		return err;

	sr = jfaudfile_read(file, data->data, samps * KWV_CHANNELS * KWV_BYTESPERSAMP);
	if (sr <= 0) {
		jfaudbuffer_free(data);
		if (sr < 0) return jfauderr_genericerror;
		return jfauderr_ok;
	}

	data->numsamples = sr / (KWV_CHANNELS * KWV_BYTESPERSAMP);
	kwvf->audiopos += sr;

	return jfauderr_ok;
}

static jfauderr rewindpcmdata(jfaud_file *file)
{
	struct kwvinfo *kwvf;

	if (file->type != jfaudfile_kwv) return jfauderr_unrecognised;

	kwvf = (struct kwvinfo *)file->readerinfo;

	kwvf->audiopos = 0;
	jfaudfile_seek(file, kwvf->audiostart, SEEK_SET);

	return jfauderr_ok;
}

static jfauderr getpcmlength(jfaud_file *file, unsigned int *samples)
{
	struct kwvinfo *kwvf;
	
	if (file->type != jfaudfile_kwv) return jfauderr_unrecognised;

	kwvf = (struct kwvinfo *)file->readerinfo;
	*samples = kwvf->audioleng / (KWV_CHANNELS * KWV_BYTESPERSAMP);

	return jfauderr_ok;
}

static jfauderr getplaytime(jfaud_file *file, double *len)
{
	struct kwvinfo *kwvf;
	
	if (file->type != jfaudfile_kwv) return jfauderr_unrecognised;

	kwvf = (struct kwvinfo *)file->readerinfo;
	*len = (double)(kwvf->audioleng / (KWV_CHANNELS * KWV_BYTESPERSAMP)) / (double)KWV_SAMPLERATE;

	return jfauderr_ok;
}


jfaud_filereader reader_kwv = {
	"KWV Container",
	jfaudfile_kwv,
	jfaudfiledata_pcm,
	
	init,
	uninit,
	testfile,
	cleanup,

	readpcmdata,
	readallpcmdata,
	rewindpcmdata,
	getpcmlength,
	getplaytime
};
