#ifndef __jfaudinternal_h__
#define __jfaudinternal_h__

#include "jfaud.h"

#include "midiseq.h"

#ifdef __cplusplus
extern "C" {
#endif

#include <fcntl.h>

#ifndef max
#define max(a,b) ((a)>(b)?(a):(b))
#endif

#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#endif

#ifdef DEBUG
#define debuglog(f,s) f(s)
#else
#define debuglog(f,s)
#endif

#ifdef __GNUC__
#define __int64 long long
#endif

struct jfaud_init_params * jfaud_getparams(void);

typedef __int64 jfaudident;

typedef enum {
	jfaudfile_unidentified = -1,
	jfaudfile_wave = 0,
	jfaudfile_voc,
	jfaudfile_vorbis,
	jfaudfile_mpeg,
	jfaudfile_kwv,

	jfaudfile_midi,
	
	jfaudfile_cda
} jfaudfiletype;

typedef enum {
	jfaudfiledata_pcm = 0,
	jfaudfiledata_midi,
	jfaudfiledata_cda
} jfaudfiledatatype;

typedef struct {
	unsigned short ref;
	jfaudident ident;
	int fh;
	jfaudfiletype type;
	void *readerinfo;
} jfaud_file;

typedef struct {
	int samplerate;
	short channels;
	short bitspersample;
	int numsamples;
	void *data;				// if this is null and albuf is 0, the buffer is null
	unsigned int albuf;			// OpenAL buffer identifier
} jfaud_buffer;

typedef struct _jfaud_bufferlink {
	struct _jfaud_bufferlink *next,*prev;
	jfaudident ident;
	int ref, age;
	jfaud_buffer buffer;
} jfaud_bufferlink;

typedef struct {
	char used;
	unsigned long age;
	jfaud_file *data;			// NULL for non-streamed channels
	jfaudident ident;			// if bit 32 is set, this is a doublespeed-hacked channel
	char threed,stream,loop;
	
	float volume;
	float pitch;
	int fx;
	float posx,posy,posz;
	float velx,vely,velz;
	char posrel;				// 0 = listener-relative, 1 = 3d
	
	int priority;
	jfaudplaymode playmode;
	int outhandle;				// handle issued by the output device
} jfaud_channel;

typedef struct {
	const char *name;			// textual name
	jfaudfiletype id;			// filetype id
	jfaudfiledatatype datatype;		// data disposition

	jfauderr (*init)(void);			// initialise the reader
	jfauderr (*uninit)(void);		// uninitialise the reader
	jfauderr (*testfile)(jfaud_file *, const char *sub);	// identifies a file and prepares to read it
	jfauderr (*cleanup)(jfaud_file *, char *noclose);	// cleans up a file

	// pcm audio readers must supply these functions
	jfauderr (*readpcmdata)(jfaud_file *, jfaud_buffer *data, int samples);	// read some samples from the cursor posn
	jfauderr (*readallpcmdata)(jfaud_file *, jfaud_buffer *data);	// read all samples from the cursor posn to end
	jfauderr (*rewindpcmdata)(jfaud_file *);	// rewind the cursor to the beginning
	jfauderr (*getpcmlength)(jfaud_file *, unsigned int *samples);	// playback length in samples

	jfauderr (*getplaytime)(jfaud_file *, double *len);	// playback time of sample in seconds
} jfaud_filereader;


extern jfaud_filereader reader_wave;
extern jfaud_filereader reader_voc;
extern jfaud_filereader reader_vorbis;
extern jfaud_filereader reader_midi;
extern jfaud_filereader reader_mpeg;
extern jfaud_filereader reader_kwv;

#define numfilereaders 6
extern jfaud_filereader *filereaders[numfilereaders];
extern char filereaderspresent[numfilereaders];

extern char simulateaudiolib;

#define reserved_priority	(-1)

#define stream_samples		10240

jfauderr jfaudfile_init(void);
jfauderr jfaudfile_uninit(void);
jfauderr jfaudfile_open(const char *fn, jfaud_file **file);
jfauderr jfaudfile_close(jfaud_file *file);
jfauderr jfaudfile_closehandle(jfaud_file *file);
int jfaudfile_read(jfaud_file *file, void *buf, unsigned len);
int jfaudfile_seek(jfaud_file *file, int len, int whence);
int jfaudfile_tell(jfaud_file *file);
jfaudident jfaud_hashfn(const unsigned char *blk);
jfauderr jfaudfile_getpcmlength(jfaud_file *file, unsigned int *samples);
jfauderr jfaudfile_getplaytime(jfaud_file *file, double *len);
void jfaudfile_dumpstatus(void);

jfauderr jfaudbuffer_alloc(jfaud_buffer *buf, int samplerate, short channels, short bitspersample, int numsamples);
jfauderr jfaudbuffer_enlarge(jfaud_buffer *buf, int numsamples);
jfauderr jfaudbuffer_free(jfaud_buffer *buf);				// don't call this on a managed buffer

// why this? because buffers can be used for temporary storage purposes, and do indeed get used for that by
// the fx units. there's no point wasting time managing temporary buffers, so only buffers that really warrant
// management, like cached single-shot sounds, are kept in memory
jfauderr jfaudbuffer_init(void);
jfauderr jfaudbuffer_uninit(void);
jfauderr jfaudbuffer_fetch(jfaudident ident, jfaud_buffer **buf);	// fetches a buffer for a non-streamed sound
jfauderr jfaudbuffer_isavailable(jfaudident ident);	// checks if a buffer is loaded
jfauderr jfaudbuffer_release(jfaudident ident);			// releases a lock on a managed buffer
jfauderr jfaudbuffer_manage(jfaudident ident, jfaud_buffer *buf);	// submits a buffer for management
jfauderr jfaudbuffer_update(void);				// ages buffers and cleans up those that are rotten

jfauderr jfaudfx_audiolib(jfaud_buffer *buf);

jfauderr jfaudchan_init(void);
jfauderr jfaudchan_uninit(void);
jfauderr jfaudchan_update(void);
	// alloc fetches the buffer from jfaudbuffer_fetch() and if it fails to get the data, returns a failure
jfauderr jfaudchan_alloc(int *chan, jfaudident ident, int priority);
	// this handles all cases where the sound is not played from a managed buffer, like with midi files, cda, etc.
jfauderr jfaudchan_allocstream(int *chan, jfaud_file *data, int priority);
	// this releases the buffer using jfaudbuffer_release() if the channel is not a stream
jfauderr jfaudchan_release(int chan);
jfauderr jfaudchan_setprops(int chan, float volume, float pitch, char loop, char threed);
jfauderr jfaudchan_setpos(int chan,
		float posx, float posy, float posz,
		float velx, float vely, float velz,
		char posrel);
jfauderr jfaudchan_getplaymode(int chan, jfaudplaymode *mode);
jfauderr jfaudchan_setplaymode(int chan, jfaudplaymode mode);

jfauderr jfaudopenal_init(void);
jfauderr jfaudopenal_uninit(void);
jfauderr jfaudopenal_allocsource(int *handle, char stream);
jfauderr jfaudopenal_releasesource(int handle);
jfauderr jfaudopenal_preparebuffer(jfaud_buffer *buf);		// this loads the sound data into an openal buffer
jfauderr jfaudopenal_releasebuffer(jfaud_buffer *buf);		// releases the AL source for this buffer
jfauderr jfaudopenal_queuedata(int handle, jfaud_buffer *buf);	// this attaches a buffer to a source
jfauderr jfaudopenal_needsdata(int handle);
jfauderr jfaudopenal_update(void);
jfauderr jfaudopenal_setsourceprops(int handle, float volume, float pitch, char loop, char threed);
jfauderr jfaudopenal_setsourcepos(int handle,
		float posx, float posy, float posz,
		float velx, float vely, float velz,
		char posrel);
jfauderr jfaudopenal_getsourceplaymode(int handle, jfaudplaymode *mode);
jfauderr jfaudopenal_setsourceplaymode(int handle, jfaudplaymode mode);
jfauderr jfaudopenal_setpos(float posx, float posy, float posz,
		float velx, float vely, float velz,
		float atx, float aty, float atz,
		float upx, float upy, float upz);

enum {
	Card_GeneralMIDI = 0,
	Card_SoundCanvas,
	Card_AWE32,
	Card_WaveBlaster,
	Card_OPL2,
	Card_MediaVision,
	Card_SoundMan16,
	Card_Adlib,
	Card_Soundscape,
	Card_GUS
};

extern int midicardtype;
jfauderr jfaudmidi_init(void);
jfauderr jfaudmidi_uninit(void);
jfauderr jfaudmidi_stop(void);
jfauderr jfaudmidi_play(midiseq *file, int pause);
jfauderr jfaudmidi_setplaymode(jfaudplaymode mode);
jfauderr jfaudmidi_getplaymode(jfaudplaymode *mode);
jfauderr jfaudmidi_setloop(int onf);

char *jsprintf(const char *fmt, ...);

#ifdef __cplusplus
}
#endif

#endif
