#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "internal.h"

typedef struct {
	WaveformFile_defn
	
	unsigned int samplerate;
	unsigned char channels, bitspersample, filler[2];

	unsigned long datastart, datalen, blocksize, datapos;
} SndAu;

#ifdef _MSC_VER
#pragma pack(1)
#endif

enum {
	aufmt_8bit_pcm = 2,
	aufmt_16bit_pcm = 3,
	aufmt_24bit_pcm = 4,
	aufmt_32bit_pcm = 5,
	aufmt_32bit_ieee = 6,
	aufmt_64bit_ieee = 7
};

typedef struct PACKED {
	unsigned char magic[4];
	unsigned long dataoffs;
	unsigned long datalen;
	unsigned long encoding;
	unsigned long samplerate;
	unsigned long channels;
} auhead;

#ifdef _MSC_VER
#pragma pack()
#endif



// SoundFileReader functions

static jfauderr init(void)
{
	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
	auhead auh;
	
	SndAu au;
	
	if (!file) return jfauderr_genericerror;
	if (file->sf) return jfauderr_genericerror;	// already been identified

	sysf_seek(&file->fh, 0, SEEK_SET);

	if (sysf_read(&file->fh, &auh, sizeof(auh)) != sizeof(auh))
		return jfauderr_notrecognised;
	auh.dataoffs   = BigLong(auh.dataoffs);
	auh.datalen    = BigLong(auh.datalen);
	auh.encoding   = BigLong(auh.encoding);
	auh.samplerate = BigLong(auh.samplerate);
	auh.channels   = BigLong(auh.channels);
	
	if (auh.magic[0] != '.' || auh.magic[1] != 's' || auh.magic[2] != 'n' || auh.magic[3] != 'd')
		return jfauderr_notrecognised;
	if (auh.dataoffs < 24)
		return jfauderr_filedamaged;

	if (auh.channels != 1 && auh.channels != 2)
		return jfauderr_notcompatible;

	memset(&au, 0, sizeof(au));
	au.type = type_wave;
	au.format = fmt_au;
	
	au.samplerate = auh.samplerate;
	au.channels = auh.channels;
	switch (auh.encoding) {
		case aufmt_8bit_pcm: au.bitspersample = 8; break;
		case aufmt_16bit_pcm: au.bitspersample = 16; break;
		default: return jfauderr_notcompatible;
	}

	au.datastart = auh.dataoffs;
	au.datalen = auh.datalen;
	au.blocksize = au.channels * au.bitspersample / 8;
	au.datapos = 0;

	sysf_seek(&file->fh, auh.dataoffs, SEEK_SET);
	if (sysf_tell(&file->fh) != (int)auh.dataoffs) return jfauderr_filedamaged;
	
	file->sf = (SoundFile *)malloc(sizeof(SndAu));
	if (!file->sf) return jfauderr_outofmemory;	// out of memory

	memcpy(file->sf, &au, sizeof(SndAu));

	return jfauderr_noerror;
}

static jfauderr unload(File *file UNUSED, char *noclose)
{
	// caller cleans up file->sf
	*noclose = 0;

	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file UNUSED, JFAudMetaInfo what UNUSED, char *buf UNUSED, int *len UNUSED)
{
	return jfauderr_notimpl;
}


// WaveformFileReader functions
static jfauderr loadsound(File *file, SndAu *au, PcmBuffer *buf, unsigned long samples, char loop)
{
	long readbytes, justrewound = 0;
	jfauderr err;
	char *rp;

	err = pcmbuffer_allocate(buf, samples, au->samplerate, au->channels, au->bitspersample);
	if (err != jfauderr_noerror) return err;
	buf->numsamples = 0;
	
	rp = (char *)buf->data;
	do {
		readbytes = sysf_read(&file->fh, rp, min(au->datalen - au->datapos, samples * au->blocksize));
		debuglog(DEBUGLVL_TRACE, jsprintf("sysf_read() = %ld",readbytes));
		if (readbytes < 0) {
			// read failures are always fatal
			pcmbuffer_free(buf);
			return jfauderr_ioerror;
		} else if (readbytes == 0) {
			if (!loop) {
				// zero bytes read is fatal only if not looping
				pcmbuffer_free(buf);
				break;
			} else {
				// otherwise we rewind
				if (justrewound) loop = 0;	// after rewinding we read 0, so break the loop
				else {
					sysf_seek(&file->fh, au->datastart, SEEK_SET);
					au->datapos = 0;
					justrewound = 1;
				}
			}
		} else {
			unsigned long i,s;
			s = readbytes / au->blocksize;
			buf->numsamples += s;

			samples -= s;
			justrewound = 0;

			if (au->bitspersample == 8) {
				unsigned char *p;
				// convert signed chars to unsigned
				p = (unsigned char *)rp;
				for (i=readbytes; i>0; i--, p++) {
					*p = (unsigned char)((signed char)*p + 0x80);
				}
#if MustSwapBig
			} else if (au->bitspersample == 16) {
				unsigned short *p;
				// byte swap shorts
				p = (unsigned short *)rp;
				for (i=readbytes/2; i>0; i--, p++) {
					*p = BigShort(*p);
				}
#endif
			}
		}

		au->datapos += readbytes;
		rp += readbytes;

	} while (samples > 0 && loop);

	if (buf->numsamples > 0)
		debuglog(DEBUGLVL_WORDY,jsprintf("Read %d sample(s)", buf->numsamples));

	return jfauderr_noerror;
}

static jfauderr readsamples(File *file, int n, PcmBuffer *buf, char loop)
{
	SndAu *au;

	au = (SndAu *)file->sf;

	if (n <= 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	return loadsound(file, au, buf, n, loop);
}

static jfauderr readallsamples(File *file, PcmBuffer *buf)
{
	SndAu *au;
	unsigned long samples;

	au = (SndAu *)file->sf;

	samples = au->datalen / au->blocksize;
	if (samples == 0 || au->datalen == 0xffffffff) {
		// if datalen == -1, we don't know how long the file actually is.
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	sysf_seek(&file->fh, au->datastart, SEEK_SET); au->datapos = 0;
	
	return loadsound(file, au, buf, samples, 0);
}

static jfauderr getplaytime(File *file, float *len)
{
	SndAu *au;

	au = (SndAu *)file->sf;

	if (au->datalen == 0xffffffff) *len = -1.0;	// unspecified file length in header
	else *len = (float)(au->datalen / au->blocksize) / (float)au->samplerate;
	
	return jfauderr_noerror;
}


WaveformFileReader wf_au = {
	"Sun SND/AU",
	init,
	uninit,

	type_wave,
	test,
	unload,
	getmetainfo,

	fmt_au,
	readsamples,
	readallsamples,
	getplaytime
};

