#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "internal.h"

typedef struct {
	WaveformFile_defn
	
	unsigned int samplerate;
	unsigned char channels, bitspersample, filler[2];

	unsigned long datastart, datalen, blocksize, datapos;

	char *title, *artist;
} RiffWav;

#ifdef _MSC_VER
#pragma pack(1)
#endif

typedef struct PACKED {
	char id[4];
	unsigned long length;
} riffchunkhead;

typedef struct PACKED {
	short formattag;
	short channels;
	long samplespersec;
	long avgbytespersec;
	short blockalign;
} wavefmt;

enum {
	wavefmt_pcm = 1
};

typedef struct PACKED {
	short bitspersample;
} wavefmtpcm;

#ifdef _MSC_VER
#pragma pack()
#endif



// SoundFileReader functions

static jfauderr init(void)
{
	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
	riffchunkhead rch;
	char gotfmt = 0, gotdata = 0;
	char type[4];
	long chunksiz = 0;
	
	RiffWav wav;
	
	memset(&wav, 0, sizeof(wav));
	wav.type = type_wave;
	wav.format = fmt_riffwav;
	
	if (!file) return jfauderr_genericerror;
	if (file->sf) return jfauderr_genericerror;	// already been identified

	sysf_seek(&file->fh, 0, SEEK_SET);

	if (sysf_read(&file->fh, &rch, sizeof(rch)) != sizeof(rch))
		return jfauderr_notrecognised;
	if (rch.id[0] != 'R' || rch.id[1] != 'I' || rch.id[2] != 'F' || rch.id[3] != 'F')
		return jfauderr_notrecognised;

	if (sysf_read(&file->fh, type, sizeof(type)) != sizeof(type))
		return jfauderr_filedamaged;
	if (type[0] != 'W' || type[1] != 'A' || type[2] != 'V' || type[3] != 'E')
		return jfauderr_notcompatible;

	while (1) {
		int r;

		r = sysf_read(&file->fh, &rch, sizeof(rch));
		if (r == 0) break;
		if (r != sizeof(rch)) return jfauderr_filedamaged;

		if (chunksiz <= 0) {
			if (rch.id[0] == 'd' && rch.id[1] == 'a' && rch.id[2] == 't' && rch.id[3] == 'a') {
				if (!gotfmt) {
					// missing the preceeding 'fmt ' chunk
					return jfauderr_notcompatible;
				} else {
					gotdata = 1;
					wav.datastart = sysf_tell(&file->fh);
					wav.datalen = rch.length;

					if (rch.length & 1) rch.length++;
					sysf_seek(&file->fh, rch.length, SEEK_CUR);
				}

			} else if (rch.id[0] == 'f' && rch.id[1] == 'm' && rch.id[2] == 't' && rch.id[3] == ' ') {
				wavefmt wf;

				gotfmt = 1;
				if (sysf_read(&file->fh, &wf, sizeof(wf)) != sizeof(wf))
					return jfauderr_filedamaged;

				wav.samplerate = wf.samplespersec;
				wav.channels = wf.channels;
				wav.blocksize = wf.blockalign;
				
				if (wf.formattag == wavefmt_pcm) {
					wavefmtpcm wfpcm;

					if (sysf_read(&file->fh, &wfpcm, sizeof(wfpcm)) != sizeof(wfpcm))
						return jfauderr_filedamaged;

					wav.bitspersample = wfpcm.bitspersample;
				} else {
					// unsupported data format
					return jfauderr_notcompatible;
				}

			} else if (rch.id[0] == 'L' && rch.id[1] == 'I' && rch.id[2] == 'S' && rch.id[3] == 'T') {
				if (sysf_read(&file->fh, type, sizeof(type)) != sizeof(type))
					return jfauderr_filedamaged;

				if (type[0] != 'I' || type[1] != 'N' || type[2] != 'F' || type[3] != 'O') {
					sysf_seek(&file->fh, rch.length-sizeof(type), SEEK_CUR);
				} else {
					chunksiz = rch.length - sizeof(type);
				}

			} else {
				// skip over unrecognised chunk
				if (rch.length & 1) rch.length++;
				sysf_seek(&file->fh, rch.length, SEEK_CUR);
			}
		} else {
			if (rch.length & 1) rch.length++;
			if (rch.id[0] == 'I' && rch.id[1] == 'A' && rch.id[2] == 'R' && rch.id[3] == 'T') {
				wav.artist = (char *)malloc(rch.length);
				if (!wav.artist) return jfauderr_outofmemory;
				
				if ((unsigned long)sysf_read(&file->fh, wav.artist, rch.length) != rch.length) {
					if (gotdata) break;
					return jfauderr_filedamaged;
				}
				wav.artist[ rch.length - 1 ] = 0;
				debuglog(DEBUGLVL_WORDY,jsprintf("artist=%s", wav.artist));
			} else if (rch.id[0] == 'I' && rch.id[1] == 'N' && rch.id[2] == 'A' && rch.id[3] == 'M') {
				wav.title = (char *)malloc(rch.length);
				if (!wav.title) return jfauderr_outofmemory;
				
				if ((unsigned long)sysf_read(&file->fh, wav.title, rch.length) != rch.length) {
					if (gotdata) break;
					return jfauderr_filedamaged;
				}
				wav.title[ rch.length - 1 ] = 0;
				debuglog(DEBUGLVL_WORDY,jsprintf("title=%s", wav.title));
			} else {
				sysf_seek(&file->fh, rch.length, SEEK_CUR);
			}
			chunksiz -= rch.length;
		}
	}
	
	wav.datapos = 0;
	sysf_seek(&file->fh, wav.datastart, SEEK_SET);
	
	file->sf = (SoundFile *)malloc(sizeof(RiffWav));
	if (!file->sf) return jfauderr_outofmemory;	// out of memory

	memcpy(file->sf, &wav, sizeof(RiffWav));

	return jfauderr_noerror;
}

static jfauderr unload(File *file, char *noclose)
{
	RiffWav *wav;

	wav = (RiffWav *)file->sf;

	if (wav->artist) free(wav->artist);
	if (wav->title) free(wav->title);
	
	// caller cleans up file->sf
	*noclose = 0;

	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file, JFAudMetaInfo what, char *buf, int *len)
{
	RiffWav *wav;
	char *t;

	wav = (RiffWav *)file->sf;
	
	switch (what) {
		case JFAudMetaInfo_title: t = wav->title; break;
		case JFAudMetaInfo_artist: t = wav->artist; break;
		default: return jfauderr_genericerror;
	}

	if (!t) return jfauderr_genericerror;
	*len = min(*len, (int)strlen(t));
	memcpy(buf, t, *len);
	
	return jfauderr_noerror;
}


// WaveformFileReader functions
static jfauderr loadsound(File *file, RiffWav *wav, PcmBuffer *buf, unsigned long samples, char loop)
{
	long readbytes, justrewound = 0;
	jfauderr err;
	char *rp;

	err = pcmbuffer_allocate(buf, samples, wav->samplerate, wav->channels, wav->bitspersample);
	if (err != jfauderr_noerror) return err;
	buf->numsamples = 0;
	
	rp = (char *)buf->data;
	do {
		readbytes = sysf_read(&file->fh, rp, min(wav->datalen - wav->datapos, samples * wav->blocksize));
		debuglog(DEBUGLVL_TRACE, jsprintf("sysf_read() = %ld",readbytes));
		if (readbytes < 0) {
			// read failures are always fatal
			pcmbuffer_free(buf);
			return jfauderr_ioerror;
		} else if (readbytes == 0) {
			if (!loop) {
				// zero bytes read is fatal only if not looping
				pcmbuffer_free(buf);
				break;
			} else {
				// otherwise we rewind
				if (justrewound) loop = 0;	// after rewinding we read 0, so break the loop
				else {
					sysf_seek(&file->fh, wav->datastart, SEEK_SET);
					wav->datapos = 0;
					justrewound = 1;
				}
			}
		} else {
			unsigned long s;
			s = readbytes / wav->blocksize;
			buf->numsamples += s;

			samples -= s;
			justrewound = 0;

#if MustSwapLittle
			if (wav->bitspersample == 16) {
				unsigned long i;
				unsigned short *p;
				// byte swap shorts
				p = (unsigned short *)rp;
				for (i=readbytes/2; i; i--, p++) {
					*p = LittleShort(*p);
				}
			}
#endif
		}

		wav->datapos += readbytes;
		rp += readbytes;

	} while (samples > 0 && loop);

	if (buf->numsamples > 0)
		debuglog(DEBUGLVL_WORDY,jsprintf("Read %d sample(s)", buf->numsamples));

	return jfauderr_noerror;
}

static jfauderr readsamples(File *file, int n, PcmBuffer *buf, char loop)
{
	RiffWav *wav;

	wav = (RiffWav *)file->sf;

	if (n <= 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	return loadsound(file, wav, buf, n, loop);
}

static jfauderr readallsamples(File *file, PcmBuffer *buf)
{
	RiffWav *wav;
	unsigned long samples;

	wav = (RiffWav *)file->sf;

	samples = wav->datalen / wav->blocksize;
	if (samples == 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	sysf_seek(&file->fh, wav->datastart, SEEK_SET); wav->datapos = 0;
	
	return loadsound(file, wav, buf, samples, 0);
}

static jfauderr getplaytime(File *file, float *len)
{
	RiffWav *wav;

	wav = (RiffWav *)file->sf;

	*len = (float)(wav->datalen / wav->blocksize) / (float)wav->samplerate;
	
	return jfauderr_noerror;
}


WaveformFileReader wf_riffwav = {
	"RIFF Wave",
	init,
	uninit,

	type_wave,
	test,
	unload,
	getmetainfo,

	fmt_riffwav,
	readsamples,
	readallsamples,
	getplaytime
};

